<?php
namespace Zeedhi\ZhuLog\Listeners;

use Zeedhi\Framework\Events\PreDispatch\Listener as PreDispatchListener;
use Zeedhi\Framework\Cache\Type\MemcachedImpl;
use Zeedhi\Framework\Events\AbstractEvent;
use Zeedhi\Framework\Events\Listener;
use Doctrine\DBAL\Connection;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\HTTP;

use Zeedhi\ZhuLog\Controller\ZhuLogConfiguration;
use Zeedhi\ZhuLog\Logger\Logger;

class OnRequest extends PreDispatchListener{

    const PREFIX = 'Zeedhi\Framework\DTO\Request\\';

    /** @var Logger $logger The logger object. */
    protected $logger;
    /** @var MemcachedImpl $memcached The Memcached object. */
    protected $memcached;
    /** @var Connection $connection Connection object. */
    protected $connection;
    /** @var AbstractEvent $postDispatchEvent Post Dispatch Event object. */
    protected $postDispatchEvent;
    /** @var Listener $beforeResponseListener Before Response Listener object. */
    protected $beforeResponseListener;
    /** @var AbstractEvent $onExceptionEvent On Exception Event object. */
    protected $onExceptionEvent;
    /** @var Listener $onExceptionListener On Exception Listener object. */
    protected $onExceptionListener;
    /** @var string $baseUri Defines the base of the project for http requests. */
    private $baseUri;
    /** @var boolean $skip Defines if the event dealing shall be skipped or not. */
    private $skip = false;

    /**
     * Constructor...
     *
     * @param Logger        $logger                 The logger object.
     * @param MemcachedImpl $memcached              The Memcached object.
     * @param Connection    $connection             Connection object.
     * @param AbstractEvent $postDispatchEvent      Post Dispach Event object.
     * @param Listener      $beforeResponseListener Post Dispatch Listener object.
     * @param AbstractEvent $onExceptionEvent       On Exception Event object.
     * @param Listener      $onExceptionListener    On Exception Listener object.
     * @param string[]      $skippedRoutes          Array of skipped routes.
     */
    public function __construct(
        Logger $logger,
        MemcachedImpl $memcached,
        Connection $connection,
        AbstractEvent $postDispatchEvent,
        Listener $beforeResponseListener,
        AbstractEvent $onExceptionEvent,
        Listener $onExceptionListener,
        string $baseUri,
        array $skippedRoutes = []
    ){
        $this->logger = $logger;
        $this->memcached = $memcached;
        $this->connection = $connection;
        $this->postDispatchEvent = $postDispatchEvent;
        $this->beforeResponseListener = $beforeResponseListener;
        $this->onExceptionEvent = $onExceptionEvent;
        $this->onExceptionListener = $onExceptionListener;

        $this->baseUri = $baseUri;

        $this->memcached->getMemcached()->addServer('127.0.0.1', 11211);

        $this->preventUselessLog($skippedRoutes);
    }

    /**
     * Checks if is FilterData request.
     *
     * @param string $type The Request type.
     *
     * @return boolean
     */
    private function isFilter($type){
        return $type === Request::TYPE_FILTER;
    }

    /**
     * Checks if the actual route should be skipped.
     *
     * @param string   $uri           The actual called uri.
     * @param string[] $skippedRoutes The routes to be skipped.
     *
     * @return boolean
     */
    private function toSkip($uri, array $skippedRoutes){
        $this->uri = $uri;

        $dontSkip = empty(array_filter(
            $skippedRoutes,
            function($route){
                return strrpos($this->uri, $route) !== false;
            }
        ));

        return !$dontSkip;
    }

    /**
     * Checks if zhulog status is cached.
     *
     * @param string $productId      The product ID.
     * @param string $organizationId The organization ID.
     * @param string $userId         The user ID.
     *
     * @return boolean
     */
    private function isCached($productId, $organizationId, $userId){
        $cached = $this->memcached->contains($productId);

        if(!$cached){
            $cached = $this->populateCache($productId, $organizationId, $userId);
        }

        return $cached;
    }

    /**
     * Checks if the user is allowed to log.
     *
     * @param string $productId      The product ID.
     * @param string $organizationId The organization ID.
     * @param string $userId         The user ID.
     *
     * @return boolean
     */
    private function isAllowed($productId, $organizationId, $userId, $cached){
        if(!$cached) return false;

        if($this->memcached->fetch($productId) === 'Y') return true;

        $key = "allowedUsers_{$productId}";
        $data = null;
        if($this->memcached->contains($key)){
            $data = $this->memcached->fetch($key);
            if(strlen($data) === 0) return false;

            $allowedUsers = json_decode($data, true);
            return empty($allowedUsers) || $this->inArray($userId, $allowedUsers);
        }
    }

    /**
     * Returns TRUE if user id is found at allowed users array.
     */
    private function inArray($userId, $allowedUsers){
        foreach($allowedUsers as $allowedUser){
            if($userId === $allowedUser['USER_ID']) return true;
        }
    }

    /**
     * Prevents useless event calls.
     *
     * @param string[] $toSkip The routes to skip.
     */
    private function preventUselessLog(array $toSkip){
        $request = HTTP\Request::initfromGlobals();
        $type = $request->getRequestType();
        $uri = $this->cleanupUri($request->getRequestUri());
        $cached = false;

        list($userId, $organizationId, $productId) = $this->logger->getUserData();

        if(!$this->toSkip($uri, $toSkip)){
            $cached = $this->isCached($productId, $organizationId, $userId);
            $cached = $cached && $this->memcached->fetch($productId);
        }
        if(
            !$cached ||
            $this->isFilter($type) ||
            !$this->isAllowed($productId, $organizationId, $userId, $cached)
        ){
            $this->skip = true;
            $this->unsetEvents();
        }
    }

    /**
     * Unsets BeforeResponse Listener instance
     * from PostDipatch Event array of listeners.
     */
    private function unsetEvents(){
        $this->postDispatchEvent->removeListener($this->beforeResponseListener);
        $this->onExceptionEvent->removeListener($this->onExceptionListener);
    }

    /**
     * {@inheritdoc}
     */
    public function preDispatch(Request $request){
        if($this->skip) return;

        $uri = $request->getRoutePath();
        $method = $request->getMethod();
        $requestType = "Empty";

        switch(get_class($request)){
            case self::PREFIX.Request::TYPE_DATA_SET:
                $requestType = Request::TYPE_DATA_SET;
                break;
            case self::PREFIX.Request::TYPE_ROW:
                $requestType = Request::TYPE_ROW;
                break;
        }

        $this->logger->logRequest($uri, $method, $requestType);
    }

    /**
     * Populates memcached if table has data.
     *
     * @param string $productId      The product ID.
     * @param string $organizationId The organization ID.
     * @param string $userId         The user ID.
     *
     * @return boolean
     */
    private function populateCache($productId, $organizationId, $userId){
        $prod = ZhuLogConfiguration::buildSQL(
            $this->connection,
            ZhuLogConfiguration::ZH_LOG_PROD,
            $productId,
            $organizationId,
            $userId
        );
        $prodfetch = $prod->execute()->fetch();
        $result = $prodfetch['IS_LOG_ACTIVE'] === 'Y';
        $allowAll = $prodfetch['ALL_TOGGLED'];

        if($result){
            $this->memcached->save($productId, $allowAll);

            if(!$allowAll){
                $select = ZhuLogConfiguration::buildSQL(
                    $this->connection,
                    ZhuLogConfiguration::ZH_LOG_USER,
                    $productId,
                    $organizationId,
                    $userId
                );
                $fetch = $select->execute()->fetchAll();

                $this->memcached->save("allowedUsers_{$productId}", json_encode($fetch));
            }
        }

        return $result;
    }

    /**
     * Return a URI cleanup
     *
     * @param $requestURI
     *
     * @return mixed
     */
    protected function cleanupUri($requestURI) {
        return $this->removeParameters($this->removeBaseUri($requestURI));
    }

    /**
     * @param $requestURI
     *
     * @return mixed
     */
    protected function removeBaseUri($requestURI) {
        $this->baseUri = rtrim($this->baseUri, ' /');
        return preg_replace('#(.*)(' . $this->baseUri . ')#', '', $requestURI);
    }

    /**
     * @param $uri
     *
     * @return mixed
     */
    protected function removeParameters($uri) {
        $uris = explode("?", $uri);
        return $uris[0];
    }
}